import json
import os
import shutil
import subprocess
import traceback
from pathlib import Path
from typing import List

import questionary
import typer
from rich import print as rprint
from rich.progress import Progress, SpinnerColumn, TextColumn

import skyplane.api.usage
from skyplane import compute
from skyplane.cli.impl.common import print_header
from skyplane.api.usage import UsageClient, UsageStatsStatus
from skyplane.config import SkyplaneConfig
from skyplane.config_paths import aws_config_path, gcp_config_path, config_path, ibmcloud_config_path
from skyplane.utils import logger


def load_aws_config(config: SkyplaneConfig, non_interactive: bool = False) -> SkyplaneConfig:
    try:
        import boto3
    except ImportError:
        config.aws_enabled = False
        typer.secho("    AWS support disabled because boto3 is not installed. Run `pip install 'skyplane[aws]'`.", fg="red", err=True)
        return config
    if non_interactive or typer.confirm("    Do you want to configure AWS support in Skyplane?", default=True):
        session = boto3.Session()
        credentials_session = session.get_credentials()
        if credentials_session is None:
            config.aws_enabled = False
        else:
            credentials_frozen = credentials_session.get_frozen_credentials()
            if credentials_frozen.access_key is None or credentials_frozen.secret_key is None:
                config.aws_enabled = False
            else:
                config.aws_enabled = True

        auth = compute.AWSAuthentication(config=config)
        if config.aws_enabled:
            typer.secho(
                f"    Loaded AWS credentials from the AWS CLI [IAM access key ID: ...{credentials_frozen.access_key[-6:]}]", fg="blue"
            )
            config.aws_enabled = True
            auth.save_region_config(config)
            typer.secho(f"    AWS region config file saved to {aws_config_path}", fg="blue")
            return config
        else:
            typer.secho(
                "    AWS credentials not found in boto3 session, please use the AWS CLI to set them via `aws configure`", fg="red", err=True
            )
            typer.secho("    https://docs.aws.amazon.com/cli/latest/userguide/cli-chap-getting-started.html", fg="red", err=True)
            typer.secho("    Disabling AWS support", fg="blue")
            if auth is not None:
                auth.clear_region_config()
            return config
    else:
        config.aws_enabled = False
        typer.secho("    Disabling AWS support", fg="blue")
        return config


def load_cloudflare_config(config: SkyplaneConfig, non_interactive: bool = False) -> SkyplaneConfig:
    if non_interactive or typer.confirm("    Do you want to configure Cloudflare support in Skyplane?", default=True):
        if "R2_ACCESS_KEY_ID" in os.environ and "R2_SECRET_ACCESS_KEY" in os.environ:
            config.cloudflare_access_key_id = os.environ["R2_ACCESS_KEY_ID"]
            config.cloudflare_secret_access_key = os.environ["R2_SECRET_ACCESS_KEY"]
            typer.secho(f"Read Cloudflare R2 keys for enviornment variables.", fg="blue")
        else:
            config.cloudflare_access_key_id = typer.prompt("    Enter the R2 access key ID")
            config.cloudflare_secret_access_key = typer.prompt("    Enter the R2 secret access key")
    else:
        config.cloudflare_enabled = False
        typer.secho("    Disabling Cloudflare support", fg="blue")
    return config


def load_azure_config(config: SkyplaneConfig, force_init: bool = False, non_interactive: bool = False) -> SkyplaneConfig:
    def clear_azure_config(config, verbose=True):
        if verbose:
            typer.secho("    Disabling Azure support", fg="blue")
        config.azure_subscription_id = None
        config.azure_client_id = None
        config.azure_principal_id = None
        config.azure_enabled = False
        return config

    def make_role_cmds(principal_id, subscription_id):
        roles = ["Contributor", "Storage Blob Data Contributor", "Storage Account Contributor"]
        return [
            "az role assignment create --role".split(" ")
            + [role]
            + f"--assignee-object-id {principal_id} --assignee-principal-type ServicePrincipal".split(" ")
            + f"--subscription {subscription_id}".split(" ")
            + f"--scope /subscriptions/{subscription_id}".split(" ")
            for role in roles
        ]

    def run_az_cmd(cmd: List[str], ignore_error=False):
        out, err = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE).communicate()
        if err and not ignore_error:
            typer.secho(f"    Error running command: {cmd}", fg="red", err=True)
            typer.secho(f"    stdout: {out.decode('utf-8')}", fg="red", err=True)
            typer.secho(f"    stderr: {err.decode('utf-8')}", fg="red", err=True)
            return False, out, err
        return True, out, err

    if non_interactive or typer.confirm("    Do you want to configure Azure support in Skyplane?", default=True):
        if force_init:
            typer.secho("    Azure credentials will be re-initialized", fg="red", err=True)
            clear_azure_config(config, verbose=False)
        if config.azure_enabled and config.azure_subscription_id and config.azure_principal_id and config.azure_client_id:
            typer.secho("    Azure credentials already configured! To reconfigure Azure, run `skyplane init --reinit-azure`.", fg="blue")
            return config

        # check if az cli is installed
        if not shutil.which("az"):
            typer.secho(
                "    Azure CLI not found, please install it from https://docs.microsoft.com/en-us/cli/azure/install-azure-cli. \n Then login with `az login`",
                fg="red",
                err=True,
            )
            return clear_azure_config(config)

        # load credentials from environment variables or input
        defaults = {
            "client_id": os.environ.get("AZURE_CLIENT_ID") or config.azure_client_id,
            "subscription_id": os.environ.get("AZURE_SUBSCRIPTION_ID")
            or config.azure_subscription_id
            or compute.AzureAuthentication.infer_subscription_id(),
            "resource_group": os.environ.get("AZURE_RESOURCE_GROUP") or compute.AzureServer.resource_group_name,
            "umi_name": "skyplane_umi",
        }

        # check if the az CLI is installed
        out, err = subprocess.Popen("az --version".split(), stdout=subprocess.PIPE, stderr=subprocess.PIPE).communicate()
        if not out.decode("utf-8").startswith("azure-cli"):
            typer.secho(
                "    Azure CLI not found, please install it from https://docs.microsoft.com/en-us/cli/azure/install-azure-cli. \n Then login with `az login`",
                fg="red",
                err=True,
            )
            return clear_azure_config(config)

        # query list of subscriptions
        success, out, err = run_az_cmd("az account list -o json --all".split(" "))
        if not success:
            typer.secho("    Error listing Azure subscriptions", fg="red", err=True)
            return clear_azure_config(config)
        subscriptions = {}
        for sub in json.loads(out):
            if sub["state"] == "Enabled":
                subscriptions[sub["name"]] = sub["id"]
        defaults["subscription_name"] = (
            next((n for n, i in subscriptions.items() if i == defaults["subscription_id"]), None) if defaults["subscription_id"] else None
        )

        # select subscription to launch Skyplane VMs in
        if non_interactive:
            config.azure_subscription_id = defaults["subscription_id"]
        else:
            choices = {f"{name} ({id})": id for name, id in subscriptions.items()}
            default_choice = f"{defaults['subscription_name']} ({defaults['subscription_id']})" if defaults["subscription_id"] else None
            selected_choice = questionary.select(
                "Select Azure subscription to launch Skyplane VMs in:",
                choices=list(sorted(choices.keys())),
                default=default_choice,
                qmark="    ?",
                pointer="    > ",
            ).ask()
            if selected_choice is None:
                typer.secho("    No subscription selected, disabling Azure support", fg="blue")
                return clear_azure_config(config)
            config.azure_subscription_id = choices[selected_choice]

        if not config.azure_subscription_id:
            typer.secho("    Invalid Azure subscription ID", fg="red", err=True)
            return clear_azure_config(config)

        # ask user which subscriptions Skyplane should be able to read/write to
        # in order to move data to or from that storage account, they must select it
        if non_interactive:
            authorize_subscriptions_ids = [config.azure_subscription_id]
        else:
            choices = {f"{name} ({id})": id for name, id in subscriptions.items()}
            default_choice = f"{defaults['subscription_name']} ({defaults['subscription_id']})" if defaults["subscription_id"] else None
            authorize_subscription_strs = questionary.checkbox(
                "Select which Azure subscriptions that Skyplane should be able to read/write data to",
                choices=list(sorted(choices.keys())),
                default=default_choice,
                qmark="    ?",
                pointer="    > ",
            ).ask()
            if not authorize_subscription_strs:
                typer.secho(
                    "    Note: Skyplane will not be able to read/write data to any Azure subscriptions so you will not be able to use Azure storage.",
                    fg="red",
                )
                authorize_subscriptions_ids = []
            else:
                authorize_subscriptions_ids = [choices[s] for s in authorize_subscription_strs]

        if not config.azure_resource_group:
            config.azure_resource_group = (
                typer.prompt(
                    "    Enter the Azure resource group to provision Skyplane VMs in (the default should work in most cases)",
                    default=defaults["resource_group"],
                )
                if not non_interactive
                else defaults["resource_group"]
            )
        if not config.azure_umi_name:
            config.azure_umi_name = (
                typer.prompt(
                    "    Enter the name for the user managed identity that Skyplane VMs will use to access your Azure Storage Accounts (the default should work in most cases)",
                    default=defaults["umi_name"],
                )
                if not non_interactive
                else defaults["umi_name"]
            )

        enable_quota_provider_cmd = f"az provider register -n Microsoft.Quota"
        change_subscription_cmd = f"az account set --subscription {config.azure_subscription_id}"
        create_rg_cmd = f"az group create -l westus2 -n {config.azure_resource_group}"
        create_umi_cmd = f"az identity create -g {config.azure_resource_group} -n {config.azure_umi_name}"
        typer.secho(f"    I will run the following commands to create an Azure managed identity:", fg="blue")
        typer.secho(f"        $ {enable_quota_provider_cmd}", fg="yellow")
        typer.secho(f"        $ {change_subscription_cmd}", fg="yellow")
        typer.secho(f"        $ {create_rg_cmd}", fg="yellow")
        typer.secho(f"        $ {create_umi_cmd}", fg="yellow")

        with Progress(
            TextColumn("    "), SpinnerColumn(), TextColumn("Creating Skyplane managed identity{task.description}"), transient=True
        ) as progress:
            progress.add_task("", total=None)
            # NOTE: we want to run this command early on because it takes time to register the quota provider
            _ = run_az_cmd(enable_quota_provider_cmd.split(), ignore_error=True)

            cmd_success, out, err = run_az_cmd(change_subscription_cmd.split())
            if not cmd_success:
                return clear_azure_config(config)
            cmd_success, out, err = run_az_cmd(create_rg_cmd.split())
            if not cmd_success:
                return clear_azure_config(config)
            cmd_success, out, err = run_az_cmd(create_umi_cmd.split())
            if not cmd_success:
                return clear_azure_config(config)
            else:
                identity_json = json.loads(out.decode("utf-8"))
                config.azure_client_id = identity_json["clientId"]
                config.azure_principal_id = identity_json["principalId"]

        if (
            not config.azure_client_id
            or not config.azure_principal_id
            or not config.azure_subscription_id
            or not config.azure_umi_name
            or not config.azure_resource_group
        ):
            typer.secho("    Azure credentials not configured correctly, disabling Azure support.", fg="red", err=True)
            return clear_azure_config(config)

        # authorize new managed identity with Storage Blob Data Contributor and Storage Account Contributor roles to the subscription
        role_cmds = []
        for subscription_id in authorize_subscriptions_ids:
            role_cmds.extend(make_role_cmds(config.azure_principal_id, subscription_id))

        if role_cmds:
            typer.secho(
                f"    I will run the following commands to authorize the newly created Skyplane managed identity to access your storage accounts:",
                fg="blue",
            )
            for role_cmd in role_cmds:
                typer.secho(f"        $ {' '.join(role_cmd)}", fg="yellow")

            with Progress(
                TextColumn("    "),
                SpinnerColumn(),
                TextColumn("Authorizing managed identity to access storage accounts{task.description}"),
                transient=True,
            ) as progress:
                progress.add_task("", total=None)
                for role_cmd in role_cmds:
                    cmd_success, out, err = run_az_cmd(role_cmd)
                    if not cmd_success:
                        return clear_azure_config(config)
        typer.secho(
            f"    Azure managed identity created successfully! To delete it, run `az identity delete -n skyplane_umi -g skyplane`.",
            fg="green",
        )

        config.azure_enabled = True
        auth = compute.AzureAuthentication(config=config)
        with Progress(
            TextColumn("    "),
            SpinnerColumn(),
            TextColumn("Querying Azure for available regions and VM SKUs{task.description}"),
            transient=True,
        ) as progress:
            progress.add_task("", total=None)
            auth.save_region_config()
        return config
    else:
        return clear_azure_config(config)


def check_gcp_service(gcp_auth: compute.GCPAuthentication, non_interactive: bool = False):
    services = {"iam": "IAM", "compute": "Compute Engine", "storage": "Storage", "cloudresourcemanager": "Cloud Resource Manager"}
    for service, name in services.items():
        if not gcp_auth.check_api_enabled(service):
            typer.secho(f"    GCP {name} API not enabled", fg="red", err=True)
            if non_interactive or typer.confirm(f"    Do you want to enable the {name} API?", default=True):
                gcp_auth.enable_api(service)
                typer.secho(f"    Enabled GCP {name} API", fg="blue")
            else:
                return False
    return True


def load_gcp_config(config: SkyplaneConfig, force_init: bool = False, non_interactive: bool = False) -> SkyplaneConfig:
    def disable_gcp_support():
        typer.secho("    Disabling Google Cloud support", fg="blue")
        config.gcp_enabled = False
        config.gcp_project_id = None
        compute.GCPAuthentication.clear_region_config()
        return config

    if non_interactive or typer.confirm("    Do you want to configure GCP support in Skyplane?", default=True):
        if force_init:
            typer.secho("    GCP credentials will be re-initialized", fg="red", err=True)
            config.gcp_project_id = None
        elif not Path(gcp_config_path).is_file():
            typer.secho("    GCP region config missing! GCP will be reconfigured.", fg="red", err=True)
            config.gcp_project_id = None

        if config.gcp_project_id is not None:
            typer.secho("    GCP already configured! To reconfigure GCP, run `skyplane init --reinit-gcp`.", fg="blue")
            config.gcp_enabled = True
            return config

        # check if GCP is enabled
        inferred_cred, inferred_project = compute.GCPAuthentication.get_adc_credential()
        if inferred_cred is None:  # or inferred_project is None:
            typer.secho("    Default GCP credentials are not set up yet. Run `gcloud auth application-default login`.", fg="red", err=True)
            typer.secho("    https://cloud.google.com/docs/authentication/getting-started", fg="red", err=True)
            return disable_gcp_support()
        else:
            typer.secho("    GCP credentials found in GCP CLI", fg="blue")
            if non_interactive or typer.confirm("    Do you want to enable GCP support in Skyplane?", default=True):
                if not non_interactive:
                    config.gcp_project_id = typer.prompt("    Enter the GCP project ID", default=inferred_project)

                else:
                    config.gcp_project_id = inferred_project
                assert config.gcp_project_id is not None, "GCP project ID must not be None"
                config.gcp_enabled = True
                auth = compute.GCPAuthentication(config=config)
                typer.secho(f"    Using GCP service account {auth.service_account_name}", fg="blue")
                if not check_gcp_service(auth, non_interactive):
                    return disable_gcp_support()
                try:
                    auth.save_region_config()
                except Exception as e:
                    typer.secho(f"    Error saving GCP region config", fg="red", err=True)
                    typer.secho(f"    {e}\n{traceback.format_exc()}", fg="red", err=True)
                    return disable_gcp_support()
                return config
            else:
                return disable_gcp_support()
    else:
        return disable_gcp_support()


def load_ibmcloud_config(config: SkyplaneConfig, force_init: bool = False, non_interactive: bool = False) -> SkyplaneConfig:
    try:
        HOME_DIR = os.path.expanduser("~")
        CONFIG_DIR = os.path.join(HOME_DIR, ".bluemix")
        CONFIG_FILE = os.path.join(CONFIG_DIR, "ibm_credentials")

        def load_yaml_config(config_filename):
            import yaml

            try:
                with open(config_filename, "r") as config_file:
                    data = yaml.safe_load(config_file)
            except FileNotFoundError:
                data = {}

            return data

        def get_default_config_filename():
            """
            First checks .ibm_config
            then checks IBM_CONFIG_FILE environment variable
            then ~/.ibm/config
            """
            if "IBM_CONFIG_FILE" in os.environ:
                config_filename = os.environ["IBM_CONFIG_FILE"]

            elif os.path.exists(".ibm_config"):
                config_filename = os.path.abspath(".ibm_credentials")

            else:
                config_filename = CONFIG_FILE
                if not os.path.exists(config_filename):
                    return None

            return config_filename

        def load_config():
            """Load the configuration"""
            config_data = None
            config_filename = get_default_config_filename()
            if config_filename:
                config_data = load_yaml_config(config_filename)
            else:
                # throw exception
                raise Exception(f"IBM Config file not foud in {config_filename}")

            return config_data

    except ImportError:
        config.ibmcloud_enabled = False
        typer.secho(
            "    IBM Cloud support disabled because ibm_boto3 is not installed. Run `pip install skyplane[ibmcloud].`", fg="red", err=True
        )
        return config
    if non_interactive or typer.confirm("    Do you want to configure IBM Cloud support in Skyplane?", default=True):
        ibm_config = load_config()
        config.ibmcloud_useragent = ibm_config["user_agent"] if "user_agent" in ibm_config else "skyplane-ibm"

        if "iam" in ibm_config and "ibm_iam_key" in ibm_config["iam"]:
            config.ibmcloud_iam_key = ibm_config["iam"].get("ibm_iam_key")

        if "iam" in ibm_config and "iam_endpoint" in ibm_config["iam"]:
            config.ibmcloud_iam_endpoint = ibm_config["iam"].get("iam_endpoint")
        else:
            config.ibmcloud_iam_endpoint = "https://iam.cloud.ibm.com"

        if "cos" in ibm_config and "access_key" in ibm_config["cos"]:
            config.ibmcloud_access_id = ibm_config["cos"]["access_key"]

        if "cos" in ibm_config and "secret_key" in ibm_config["cos"]:
            config.ibmcloud_secret_key = ibm_config["cos"]["secret_key"]

        if "ibm" in ibm_config and "resource_group_id" in ibm_config["ibm"]:
            config.ibmcloud_resource_group_id = ibm_config["ibm"]["resource_group_id"]

        if config.ibmcloud_access_id is not None:
            config.ibmcloud_enabled = True

        auth = compute.IBMCloudAuthentication(config=config)
        if config.ibmcloud_enabled:
            typer.secho(f"    Loaded IBM Cloud credentials ", fg="blue")
            config.ibmcloud_enabled = True
            auth.save_region_config(config)
            typer.secho(f"    IBM Cloud  config file saved to {ibmcloud_config_path}", fg="blue")
            return config
        else:
            typer.secho(f"    COS credentials not found {get_default_config_filename()}", fg="red", err=True)
            typer.secho("    Disabling IBM Cloud support", fg="blue")
            if auth is not None:
                auth.clear_region_config()
            return config
    else:
        config.cos_enabled = False
        typer.secho("    Disabling IBM Cloud support", fg="blue")
        return config


def init(
    non_interactive: bool = typer.Option(False, "--non-interactive", "-y", help="Run non-interactively"),
    reinit_azure: bool = False,
    reinit_gcp: bool = False,
    reinit_ibm: bool = False,
    reinit_cloudflare: bool = False,
    disable_config_aws: bool = False,
    disable_config_azure: bool = False,
    disable_config_gcp: bool = False,
    disable_config_ibm: bool = True,  # TODO: eventuall enable IBM
    disable_config_cloudflare: bool = False,
):
    """
    It loads the configuration file, and if it doesn't exist, it creates a default one. Then it creates
    AWS, Azure, IBM and GCP region list configurations.

    :param reinit_azure: If true, will reinitialize the Azure region list and credentials
    :type reinit_azure: bool
    :param reinit_gcp: If true, will reinitialize the GCP region list and credentials
    :type reinit_gcp: bool
    :param reinit_ibm: If true, will reinitialize the IBM Cloud region list and credentials
    :type reinit_ibm: bool
    :param disable_config_aws: If true, will disable AWS configuration (may still be enabled if environment variables are set)
    :type disable_config_aws: bool
    :param disable_config_azure: If true, will disable Azure configuration (may still be enabled if environment variables are set)
    :type disable_config_azure: bool
    :param disable_config_gcp: If true, will disable GCP configuration (may still be enabled if environment variables are set)
    :type disable_config_gcp: bool
    :param disable_config_ibm: If true, will disable IBM Cloud configuration (may still be enabled if environment variables are set)
    :type disable_config_ibm: bool
    :param disable_config_cloudflare: If true, will disable Cloudflare configuration (may still be enabled if environment variables are set)
    :type disable_config_cloudflare: bool
    """
    print_header()

    if non_interactive:
        logger.warning("Non-interactive mode enabled. Automatically confirming interactive questions.")

    if config_path.exists():
        logger.debug(f"Found existing configuration file at {config_path}, loading")
        cloud_config = SkyplaneConfig.load_config(config_path)
    else:
        cloud_config = SkyplaneConfig.default_config()

    # load AWS config
    if not (reinit_azure or reinit_gcp or reinit_ibm):
        typer.secho("\n(1) Configuring AWS:", fg="yellow", bold=True)
        if not disable_config_aws:
            cloud_config = load_aws_config(cloud_config, non_interactive=non_interactive)

    # load Azure config
    if not (reinit_gcp or reinit_ibm):
        if reinit_azure:
            typer.secho("\nConfiguring Azure:", fg="yellow", bold=True)
        else:
            typer.secho("\n(2) Configuring Azure:", fg="yellow", bold=True)
        if not disable_config_azure:
            cloud_config = load_azure_config(cloud_config, force_init=reinit_azure, non_interactive=non_interactive)

    # load GCP config
    if not reinit_azure:
        if reinit_gcp:
            typer.secho("\nConfiguring GCP:", fg="yellow", bold=True)
        else:
            typer.secho("\n(3) Configuring GCP:", fg="yellow", bold=True)
        if not disable_config_gcp:
            cloud_config = load_gcp_config(cloud_config, force_init=reinit_gcp, non_interactive=non_interactive)

    # load cloudflare config
    if not reinit_cloudflare and not disable_config_cloudflare:  # TODO: fix reinit logic
        typer.secho("\n(4) Configuring Cloudflare R2:", fg="yellow", bold=True)
        if not disable_config_cloudflare:
            cloud_config = load_cloudflare_config(cloud_config, non_interactive=non_interactive)

    # load IBMCloud config
    if not disable_config_ibm and not reinit_ibm:
        # TODO: fix IBM configuration to not fail on file finding
        typer.secho("\n(4) Configuring IBM Cloud:", fg="yellow", bold=True)
        if not disable_config_ibm:
            cloud_config = load_ibmcloud_config(cloud_config, force_init=reinit_ibm, non_interactive=non_interactive)

    cloud_config.to_config_file(config_path)
    typer.secho(f"\nConfig file saved to {config_path}", fg="green")

    # Set metrics collection by default
    usage_stats_var = UsageClient.usage_stats_status()
    if usage_stats_var is UsageStatsStatus.DISABLED_EXPLICITLY:
        rprint(skyplane.api.usage.USAGE_STATS_DISABLED_MESSAGE)
    elif usage_stats_var in [UsageStatsStatus.ENABLED_BY_DEFAULT, UsageStatsStatus.ENABLED_EXPLICITLY]:
        rprint(skyplane.api.usage.USAGE_STATS_ENABLED_MESSAGE)
    else:
        raise Exception("Prompt message unknown.")
    return 0



#
# (C) Copyright Cloudlab URV 2020
# (C) Copyright IBM Corp. 2023
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import functools
import inspect
import re
import os
import paramiko
import time
import logging
import uuid
import random
from concurrent.futures import ThreadPoolExecutor

from skyplane.compute.ibmcloud.ibm_gen2.ssh_client import SSHClient
from skyplane.compute.ibmcloud.ibm_gen2.constants import COMPUTE_CLI_MSG, CACHE_DIR
from skyplane.compute.ibmcloud.ibm_gen2.utils import load_yaml_config, dump_yaml_config, delete_yaml_config

from skyplane.utils import imports

logger = logging.getLogger(__name__)

INSTANCE_START_TIMEOUT = 180 * 2
VPC_API_VERSION = "2021-09-21"


class IBMVPCBackend:
    def __init__(self, ibm_vpc_config):
        logger.debug("Creating IBM VPC client")
        self.name = "ibm_gen2"
        self.config = ibm_vpc_config

        self.vpc_name = None
        self.vpc_data = None
        self.vpc_key = None

        self.endpoint = self.config["endpoint"]
        self.region = self.endpoint.split("//")[1].split(".")[0]
        self.cache_dir = os.path.join(CACHE_DIR, self.name)
        self.vpc_data_filename = os.path.join(self.cache_dir, self.region + "_data")
        self.custom_image = self.config.get("custom_skyplane_image")

        logger.debug(f"Setting VPC endpoint to: {self.endpoint}")

        self.workers = []

        self.iam_api_key = self.config.get("iam_api_key")
        self.vpc_cli = self.create_vpc_cli()
        # authenticator = ibm_cloud_sdk_core.authenticators(self.iam_api_key, url=self.config.get("iam_endpoint"))
        # self.vpc_cli = ibm_vpc.VpcV1(VPC_API_VERSION, authenticator=authenticator)
        # self.vpc_cli.set_service_url(self.config["endpoint"] + "/v1")

        user_agent_string = "ibm_vpc_{}".format(self.config["user_agent"])
        self.vpc_cli._set_user_agent_header(user_agent_string)

        # decorate instance public methods with except/retry logic
        decorate_instance(self.vpc_cli, vpc_retry_on_except)

        msg = COMPUTE_CLI_MSG.format("IBM VPC")
        logger.info(f"{msg} - Region: {self.region}")

    @imports.inject("ibm_vpc", "ibm_cloud_sdk_core", pip_extra="ibmcloud")
    def create_vpc_cli(ibm_vpc, ibm_cloud_sdk_core, self):
        authenticator = ibm_cloud_sdk_core.authenticators(self.iam_api_key, url=self.config.get("iam_endpoint"))
        vpc_cli = ibm_vpc.VpcV1(VPC_API_VERSION, authenticator=authenticator)
        vpc_cli.set_service_url(self.config["endpoint"] + "/v1")
        return vpc_cli

    def _load_vpc_data(self):
        """
        Loads VPC data from local cache
        """
        self.vpc_data = load_yaml_config(self.vpc_data_filename)
        print("load VPC data")
        print(self.vpc_data)

        if not self.vpc_data:
            logger.debug(f"Could not find VPC cache data in {self.vpc_data_filename}")
        elif "vpc_id" in self.vpc_data:
            self.vpc_key = self.vpc_data["vpc_id"].split("-")[2]

    def _dump_vpc_data(self):
        """
        Dumps VPC data to local cache
        """
        dump_yaml_config(self.vpc_data_filename, self.vpc_data)

    @imports.inject("ibm_cloud_sdk_core", pip_extra="ibmcloud")
    def _create_vpc(ibm_cloud_sdk_core, self):
        """
        Creates a new VPC
        """
        if "vpc_id" in self.config:
            print("vpc_id in self.config")
            return

        if "vpc_id" in self.vpc_data:
            print("vpc_id in self.vpc_data")
            try:
                self.vpc_cli.get_vpc(self.vpc_data["vpc_id"])
                self.config["vpc_id"] = self.vpc_data["vpc_id"]
                self.config["security_group_id"] = self.vpc_data["security_group_id"]
                print(self.config)
                return
            except ibm_cloud_sdk_core.ApiException:
                pass

        vpc_info = None

        host_id = str(uuid.getnode())[-6:]
        iam_id = self.iam_api_key[:4].lower()
        self.vpc_name = self.config.get("vpc_name", f"skyplane-vpc-{iam_id}-{host_id}")
        logger.debug(f"Setting VPC name to: {self.vpc_name}")

        assert re.match("^[a-z0-9-:-]*$", self.vpc_name), 'VPC name "{}" not valid'.format(self.vpc_name)

        vpcs_info = self.vpc_cli.list_vpcs().get_result()
        for vpc in vpcs_info["vpcs"]:
            if vpc["name"] == self.vpc_name:
                vpc_info = vpc

        if not vpc_info:
            logger.debug(f"Creating VPC {self.vpc_name}")
            vpc_prototype = {}
            vpc_prototype["address_prefix_management"] = "auto"
            vpc_prototype["classic_access"] = False
            vpc_prototype["name"] = self.vpc_name
            vpc_prototype["resource_group"] = {"id": self.config["resource_group_id"]}

            response = self.vpc_cli.create_vpc(**vpc_prototype)
            vpc_info = response.result

        self.config["vpc_id"] = vpc_info["id"]
        self.config["security_group_id"] = vpc_info["default_security_group"]["id"]

        # Set the prefix used for the VPC resources
        self.vpc_key = self.config["vpc_id"].split("-")[2]

        deloy_ssh_rule = True
        deploy_icmp_rule = True

        sg_rule_prototype_ssh = {}
        sg_rule_prototype_ssh["direction"] = "inbound"
        sg_rule_prototype_ssh["ip_version"] = "ipv4"
        sg_rule_prototype_ssh["protocol"] = "tcp"
        sg_rule_prototype_ssh["port_min"] = 22
        sg_rule_prototype_ssh["port_max"] = 22

        sg_rule_prototype_icmp = {}
        sg_rule_prototype_icmp["direction"] = "inbound"
        sg_rule_prototype_icmp["ip_version"] = "ipv4"
        sg_rule_prototype_icmp["protocol"] = "icmp"
        sg_rule_prototype_icmp["type"] = 8

        sg_rules = self.vpc_cli.get_security_group(self.config["security_group_id"])
        for rule in sg_rules.get_result()["rules"]:
            if all(item in rule.items() for item in sg_rule_prototype_ssh.items()):
                deloy_ssh_rule = False
            if all(item in rule.items() for item in sg_rule_prototype_icmp.items()):
                deploy_icmp_rule = False

        if deloy_ssh_rule:
            self.vpc_cli.create_security_group_rule(self.config["security_group_id"], sg_rule_prototype_ssh)
        if deploy_icmp_rule:
            self.vpc_cli.create_security_group_rule(self.config["security_group_id"], sg_rule_prototype_icmp)

    @imports.inject("ibm_cloud_sdk_core", pip_extra="ibmcloud")
    def _create_ssh_key(ibm_cloud_sdk_core, self):
        if "ssh_key_id" in self.config:
            print("ssh_key_id in self.config")
            return

        if "ssh_key_id" in self.vpc_data:
            print("ssh_key_id in self.vpc_data")
            try:
                self.vpc_cli.get_key(self.vpc_data["ssh_key_id"])
                self.config["ssh_key_id"] = self.vpc_data["ssh_key_id"]
                self.config["ssh_key_filename"] = self.vpc_data["ssh_key_filename"]
                return
            except ibm_cloud_sdk_core.ApiException:
                pass

        keyname = f"skyplane-key-{str(uuid.getnode())[-6:]}"
        filename = os.path.join("~", ".ssh", f"{keyname}.id_rsa")
        key_filename = os.path.expanduser(filename)

        key_info = None

        def _get_ssh_key():
            for key in self.vpc_cli.list_keys().result["keys"]:
                if key["name"] == keyname:
                    return key

        print(key_filename)
        if not os.path.isfile(key_filename):
            logger.debug("Generating new ssh key pair")
            os.system(f'ssh-keygen -b 2048 -t rsa -f {key_filename} -q -N ""')
            logger.debug(f"SHH key pair generated: {key_filename}")
        else:
            key_info = _get_ssh_key()

        if not key_info:
            with open(f"{key_filename}.pub", "r") as file:
                ssh_key_data = file.read()
            try:
                key_info = self.vpc_cli.create_key(
                    public_key=ssh_key_data, name=keyname, type="rsa", resource_group={"id": self.config["resource_group_id"]}
                ).get_result()
            except ibm_cloud_sdk_core.ApiException as e:
                logger.warn(e.message)
                if "Key with name already exists" in e.message:
                    self.vpc_cli.delete_key(id=_get_ssh_key()["id"])
                    key_info = self.vpc_cli.create_key(
                        public_key=ssh_key_data,
                        name=keyname,
                        type="rsa",
                        resource_group={"id": self.config["resource_group_id"]},
                    ).get_result()
                else:
                    if "Key with fingerprint already exists" in e.message:
                        logger.error("Can't register an SSH key with the same fingerprint")
                    raise e  # can't continue the configuration process without a valid ssh key

        self.config["ssh_key_id"] = key_info["id"]
        self.config["ssh_key_filename"] = key_filename

    @imports.inject("ibm_cloud_sdk_core", pip_extra="ibmcloud")
    def _create_subnet(ibm_cloud_sdk_core, self):
        """
        Creates a new subnet
        """
        if "subnet_id" in self.config:
            if "subnet_id" in self.vpc_data and self.vpc_data["subnet_id"] == self.config["subnet_id"]:
                self.config["zone_name"] = self.vpc_data["zone_name"]
            else:
                resp = self.vpc_cli.get_subnet(self.config["subnet_id"])
                self.config["zone_name"] = resp.result["zone"]["name"]
            return

        if "subnet_id" in self.vpc_data:
            try:
                self.vpc_cli.get_subnet(self.vpc_data["subnet_id"])
                self.config["subnet_id"] = self.vpc_data["subnet_id"]
                self.config["zone_name"] = self.vpc_data["zone_name"]
                return
            except ibm_cloud_sdk_core.ApiException:
                pass

        subnet_name = f"skyplane-subnet-{self.vpc_key}"
        subnet_data = None

        subnets_info = self.vpc_cli.list_subnets(resource_group_id=self.config["resource_group_id"]).get_result()
        for sn in subnets_info["subnets"]:
            if sn["name"] == subnet_name:
                subnet_data = sn

        if not subnet_data:
            logger.debug(f"Creating Subnet {subnet_name}")
            subnet_prototype = {}
            subnet_prototype["zone"] = {"name": self.region + "-1"}
            subnet_prototype["ip_version"] = "ipv4"
            subnet_prototype["name"] = subnet_name
            subnet_prototype["resource_group"] = {"id": self.config["resource_group_id"]}
            subnet_prototype["vpc"] = {"id": self.config["vpc_id"]}
            subnet_prototype["total_ipv4_address_count"] = 256
            response = self.vpc_cli.create_subnet(subnet_prototype)
            subnet_data = response.result

        self.config["subnet_id"] = subnet_data["id"]
        self.config["zone_name"] = subnet_data["zone"]["name"]

    @imports.inject("ibm_cloud_sdk_core", pip_extra="ibmcloud")
    def _create_gateway(ibm_cloud_sdk_core, self):
        """
        Crates a public gateway.
        Gateway is used by private nodes for accessing internet
        """
        if "gateway_id" in self.config:
            return

        if "gateway_id" in self.vpc_data:
            try:
                self.vpc_cli.get_public_gateway(self.vpc_data["gateway_id"])
                self.config["gateway_id"] = self.vpc_data["gateway_id"]
                return
            except ibm_cloud_sdk_core.ApiException:
                pass

        gateway_name = f"skyplane-gateway-{self.vpc_key}"
        gateway_data = None

        gateways_info = self.vpc_cli.list_public_gateways().get_result()
        for gw in gateways_info["public_gateways"]:
            if gw["vpc"]["id"] == self.config["vpc_id"]:
                gateway_data = gw

        if not gateway_data:
            logger.debug(f"Creating Gateway {gateway_name}")
            gateway_prototype = {}
            gateway_prototype["vpc"] = {"id": self.config["vpc_id"]}
            gateway_prototype["zone"] = {"name": self.config["zone_name"]}
            gateway_prototype["name"] = gateway_name
            gateway_prototype["resource_group"] = {"id": self.config["resource_group_id"]}
            response = self.vpc_cli.create_public_gateway(**gateway_prototype)
            gateway_data = response.result

        self.config["gateway_id"] = gateway_data["id"]

        # Attach public gateway to the subnet
        self.vpc_cli.set_subnet_public_gateway(self.config["subnet_id"], {"id": self.config["gateway_id"]})

    def _create_floating_ip(self, disable_recycle=True):
        """
        Creates a new floating IP address
        """

        floating_ip_data = None
        if not disable_recycle:
            floating_ips_info = self.vpc_cli.list_floating_ips().get_result()
            for fip in floating_ips_info["floating_ips"]:
                if fip["name"].startswith("skyplane") and fip["status"] == "available":
                    floating_ip_data = fip

        if not floating_ip_data:
            floating_ip_name = f"skyplane-{str(uuid.uuid1())[-4:]}-{str(random.randint(1000,9999))}"
            logger.debug(f"Creating floating IP {floating_ip_name}")
            floating_ip_prototype = {}
            floating_ip_prototype["name"] = floating_ip_name
            floating_ip_prototype["zone"] = {"name": self.config["zone_name"]}
            floating_ip_prototype["resource_group"] = {"id": self.config["resource_group_id"]}
            response = self.vpc_cli.create_floating_ip(floating_ip_prototype)
            floating_ip_data = response.result

        floating_ip = floating_ip_data["address"]
        floating_ip_id = floating_ip_data["id"]
        return floating_ip, floating_ip_id

    def add_ips_to_security_group(self, ip_address):
        for ip in ip_address:
            ip_rule = {}
            ip_rule["direction"] = "inbound"
            ip_rule["ip_version"] = "ipv4"
            ip_rule["protocol"] = "all"
            remote = {}
            remote["address"] = ip
            ip_rule["remote"] = remote

            deploy_ip_rule = True
            sg_rules = self.vpc_cli.get_security_group(self.config["security_group_id"])
            for rule in sg_rules.get_result()["rules"]:
                if all(item in rule.items() for item in ip_rule.items()):
                    deploy_ip_rule = False

            if deploy_ip_rule:
                logger.debug(f"About to create ip rule for {ip}")
                try:
                    self.vpc_cli.create_security_group_rule(self.config["security_group_id"], ip_rule)
                    logger.debug(f"Created rule for {ip_rule}")
                except Exception as e:
                    raise e

    def create_vpc_instance(self, public=True):
        """
        Creates the master VM insatnce
        """
        if "image_id" not in self.config and "image_id" in self.vpc_data:
            self.config["image_id"] = self.vpc_data["image_id"]

        if "image_id" not in self.config:
            for image in self.vpc_cli.list_images().result["images"]:
                if "ubuntu-22" in image["name"]:
                    self.config["image_id"] = image["id"]

        name = f"skyplane-ibm-vsi-{self.vpc_key}-{str(uuid.uuid4().hex[:8])}"
        vsi = IBMVPCInstance(name, self.config, self.vpc_cli, public)

        floating_ip, floating_ip_id = self._create_floating_ip()
        if public:
            vsi.public_ip = floating_ip
            vsi.floating_ip_id = floating_ip_id

        vsi.instance_id = None
        vsi.profile_name = self.config["master_profile_name"]
        vsi.delete_on_dismantle = False
        vsi.ssh_credentials.pop("password")

        instance_data = vsi.get_instance_data()
        if instance_data:
            vsi.private_ip = instance_data["primary_network_interface"]["primary_ipv4_address"]
            vsi.instance_id = instance_data["id"]

        instance_id = vsi.create(check_if_exists=True)
        if not vsi.is_ready():
            instance_id = vsi.create(check_if_exists=True)
            vsi.wait_ready()
            self.workers.append(vsi)
            instance_data = vsi.get_instance_data()

        return instance_id, vsi

    def init(self):
        """
        Initialize the VPC
        """
        logger.debug(f"Initializing IBM VPC backend ")

        self._load_vpc_data()

        self.vpc_data = {}

        # Create the VPC if not exists
        self._create_vpc()
        # Create the ssh key pair if not exists
        self._create_ssh_key()
        # Create a new subnaet if not exists
        self._create_subnet()
        # Create a new gateway if not exists
        self._create_gateway()

        self.vpc_data = {
            "instance_id": "0af1",
            "vpc_id": self.config["vpc_id"],
            "subnet_id": self.config["subnet_id"],
            "security_group_id": self.config["security_group_id"],
            "gateway_id": self.config["gateway_id"],
            "zone_name": self.config["zone_name"],
            "ssh_key_id": self.config["ssh_key_id"],
            "ssh_key_filename": self.config["ssh_key_filename"],
        }
        self._dump_vpc_data()

    @imports.inject("ibm_cloud_sdk_core", pip_extra="ibmcloud")
    def _delete_vm_instances(ibm_cloud_sdk_core, self, all=False):
        """
        Deletes all VM instances in the VPC
        """
        msg = f"Deleting all SkyPlane worker VMs in {self.vpc_name}" if self.vpc_name else "Deleting all SkyPlane worker VMs"
        logger.info(msg)

        def delete_instance(instance_info):
            ins_name, ins_id = instance_info
            try:
                logger.info(f"Deleting instance {ins_name}")
                self.vpc_cli.delete_instance(ins_id)
            except ibm_cloud_sdk_core.ApiException as err:
                if err.code == 404:
                    pass
                else:
                    raise err

        vms_prefixes = ("skyplane-ibm") if all else ("skyplane-ibm",)

        def get_instances():
            instances = set()
            instances_info = self.vpc_cli.list_instances().get_result()
            for ins in instances_info["instances"]:
                if ins["name"].startswith(vms_prefixes):
                    instances.add((ins["name"], ins["id"]))
            return instances

        deleted_instances = set()
        while True:
            instances_to_delete = set()
            for ins_to_delete in get_instances():
                if ins_to_delete not in deleted_instances:
                    instances_to_delete.add(ins_to_delete)

            if instances_to_delete:
                with ThreadPoolExecutor(len(instances_to_delete)) as executor:
                    executor.map(delete_instance, instances_to_delete)
                deleted_instances.update(instances_to_delete)
            else:
                break

        # Wait until all instances are deleted
        while get_instances():
            time.sleep(1)

    @imports.inject("ibm_cloud_sdk_core", pip_extra="ibmcloud")
    def _delete_subnet(ibm_cloud_sdk_core, self):
        """
        Deletes all VM instances in the VPC
        """
        subnet_name = f"skyplane-subnet-{self.vpc_key}"
        if "subnet_id" not in self.vpc_data:
            subnets_info = self.vpc_cli.list_subnets().get_result()

            for subn in subnets_info["subnets"]:
                if subn["name"] == subnet_name:
                    self.vpc_data["subnet_id"] = subn["id"]

        if "subnet_id" in self.vpc_data:
            logger.info(f"Deleting subnet {subnet_name}")

            try:
                self.vpc_cli.unset_subnet_public_gateway(self.vpc_data["subnet_id"])
            except ibm_cloud_sdk_core.ApiException as err:
                if err.code == 404 or err.code == 400:
                    logger.debug(err)
                else:
                    raise err

            try:
                self.vpc_cli.delete_subnet(self.vpc_data["subnet_id"])
            except ibm_cloud_sdk_core.ApiException as err:
                if err.code == 404 or err.code == 400:
                    logger.debug(err)
                else:
                    raise err

    @imports.inject("ibm_cloud_sdk_core", pip_extra="ibmcloud")
    def _delete_gateway(ibm_cloud_sdk_core, self):
        """
        Deletes the public gateway
        """
        gateway_name = f"skyplane-gateway-{self.vpc_key}"
        if "gateway_id" not in self.vpc_data:
            gateways_info = self.vpc_cli.list_public_gateways().get_result()

            for gw in gateways_info["public_gateways"]:
                if ["name"] == gateway_name:
                    self.vpc_data["gateway_id"] = gw["id"]

        if "gateway_id" in self.vpc_data:
            logger.info(f"Deleting gateway {gateway_name}")
            try:
                self.vpc_cli.delete_public_gateway(self.vpc_data["gateway_id"])
            except ibm_cloud_sdk_core.ApiException as err:
                if err.code == 404 or err.code == 400:
                    logger.debug(err)
                else:
                    raise err

    @imports.inject("ibm_cloud_sdk_core", pip_extra="ibmcloud")
    def _delete_ssh_key(ibm_cloud_sdk_core, self):
        """
        Deletes the ssh key
        """
        keyname = f"skyplane-key-{str(uuid.getnode())[-6:]}"
        filename = os.path.join("~", ".ssh", f"{keyname}.id_rsa")
        key_filename = os.path.expanduser(filename)

        if os.path.isfile(key_filename):
            os.remove(key_filename)
        if os.path.isfile(f"{key_filename}.pub"):
            os.remove(f"{key_filename}.pub")

        if "ssh_key_id" not in self.vpc_data:
            for key in self.vpc_cli.list_keys().result["keys"]:
                if key["name"] == keyname:
                    self.vpc_data["ssh_key_id"] = key["id"]

        if "ssh_key_id" in self.vpc_data:
            logger.info(f"Deleting SSH key {keyname}")
            try:
                self.vpc_cli.delete_key(id=self.vpc_data["ssh_key_id"])
            except ibm_cloud_sdk_core.ApiException as err:
                if err.code == 404 or err.code == 400:
                    logger.debug(err)
                else:
                    raise err

    @imports.inject("ibm_cloud_sdk_core", pip_extra="ibmcloud")
    def _delete_vpc(ibm_cloud_sdk_core, self):
        """
        Deletes the VPC
        """
        if "vpc_id" not in self.vpc_data:
            vpcs_info = self.vpc_cli.list_vpcs().get_result()
            for vpc in vpcs_info["vpcs"]:
                if vpc["name"] == self.vpc_name:
                    self.vpc_data["vpc_id"] = vpc["id"]

        if "vpc_id" in self.vpc_data:
            logger.info(f'Deleting VPC {self.vpc_data["vpc_id"]}')
            try:
                self.vpc_cli.delete_vpc(self.vpc_data["vpc_id"])
            except ibm_cloud_sdk_core.ApiException as err:
                if err.code == 404 or err.code == 400:
                    logger.debug(err)
                else:
                    raise err

    def clean(self, all=False):
        """
        Clean all the backend resources
        The gateway public IP and the floating IP are never deleted
        """
        logger.debug("Cleaning IBM VPC resources")

        self._delete_vm_instances(all=all)

        if all:
            self._load_vpc_data()
            self._delete_subnet()
            self._delete_gateway()
            self._delete_ssh_key()
            self._delete_vpc()

            delete_yaml_config(self.vpc_data_filename)

    def clear(self, job_keys=None):
        """
        Delete all the workers
        """
        # clear() is automatically called after get_result()
        self.dismantle(include_master=False)

    def dismantle(self, include_master=True):
        """
        Stop all worker VM instances
        """
        if len(self.workers) > 0:
            with ThreadPoolExecutor(min(len(self.workers), 48)) as ex:
                ex.map(lambda worker: worker.stop(), self.workers)
            self.workers = []

    def get_instance(self, name, **kwargs):
        """
        Returns a VM class instance.
        Does not creates nor starts a VM instance
        """
        instance = IBMVPCInstance(name, self.config, self.vpc_cli)

        for key in kwargs:
            if hasattr(instance, key):
                setattr(instance, key, kwargs[key])

        return instance

    def get_node_status(self, node_id):
        """returns True if a node is either not recorded or not in any valid status."""
        try:
            node = self.vpc_cli.get_instance(node_id).get_result()
            return node["status"]
        except Exception as e:
            raise e


class IBMVPCInstance:
    def __init__(self, name, ibm_vpc_config, ibm_vpc_client=None, public=False):
        """
        Initialize a IBMVPCInstance instance
        VMs can have master role, this means they will have a public IP address
        """
        self.name = name.lower()
        self.config = ibm_vpc_config

        self.delete_on_dismantle = self.config["delete_on_dismantle"]
        self.profile_name = self.config["worker_profile_name"]

        self.vpc_cli = None
        self.vpc_cli = ibm_vpc_client or self._create_vpc_client()
        self.public = public

        self.ssh_client = None
        self.instance_id = None
        self.instance_data = None
        self.private_ip = None
        self.public_ip = None
        self.floating_ip_id = None
        self.home_dir = "/root"

        self.ssh_credentials = {
            "username": self.config["ssh_username"],
            "password": self.config["ssh_password"],
            "key_filename": self.config.get("ssh_key_filename", "~/.ssh/id_rsa"),
        }
        self.validated = False

    def __str__(self):
        ip = self.public_ip if self.public else self.private_ip

        if ip is None or ip == "0.0.0.0":
            return f"VM instance {self.name}"
        else:
            return f"VM instance {self.name} ({ip})"

    @imports.inject("ibm_vpc", "ibm_cloud_sdk_core", pip_extra="ibmcloud")
    def _create_vpc_client(ibm_vpc, ibm_cloud_sdk_core, self):
        """
        Creates an IBM VPC python-sdk instance
        """
        authenticator = ibm_cloud_sdk_core.authenticators.IAMAuthenticator(
            self.config.get("iam_api_key"), url=self.config.get("iam_endpoint")
        )
        ibm_vpc_client = ibm_vpc.VpcV1(VPC_API_VERSION, authenticator=authenticator)
        ibm_vpc_client.set_service_url(self.config["endpoint"] + "/v1")

        # decorate instance public methods with except/retry logic
        decorate_instance(self.vpc_cli, vpc_retry_on_except)

        return ibm_vpc_client

    def get_ssh_client(self):
        """
        Creates an ssh client against the VM only if the Instance is the master
        """

        if not self.validated and self.public and self.instance_id:
            # validate that private ssh key in ssh_credentials is a pair of public key on instance
            key_filename = self.ssh_credentials["key_filename"]
            key_filename = os.path.abspath(os.path.expanduser(key_filename))

            if not os.path.exists(key_filename):
                raise Exception(f"Private key file {key_filename} doesn't exist")

            initialization_data = self.vpc_cli.get_instance_initialization(self.instance_id).get_result()

            private_res = paramiko.RSAKey(filename=key_filename).get_base64()
            names = []
            for k in initialization_data["keys"]:
                public_res = self.vpc_cli.get_key(k["id"]).get_result()["public_key"].split(" ")[1]
                if public_res == private_res:
                    self.validated = True
                    break
                else:
                    names.append(k["name"])

            if not self.validated:
                raise Exception(f"No public key from keys: {names} on master {self} not a pair for private ssh key {key_filename}")

        if self.private_ip or self.public_ip:
            if not self.ssh_client:
                self.ssh_client = SSHClient(self.public_ip or self.private_ip, self.ssh_credentials)

        return self.ssh_client

    def del_ssh_client(self):
        """
        Deletes the ssh client
        """
        if self.ssh_client:
            try:
                self.ssh_client.close()
            except Exception:
                pass
            self.ssh_client = None

    def is_ready(self):
        """
        Checks if the VM instance is ready to receive ssh connections
        """
        login_type = "password" if "password" in self.ssh_credentials and not self.public else "publickey"
        try:
            self.get_ssh_client().run_remote_command("id")
        except Exception as err:
            logger.debug(f"SSH to {self.public_ip if self.public else self.private_ip} failed ({login_type}): {err}")
            self.del_ssh_client()
            return False
        return True

    def wait_ready(self, timeout=INSTANCE_START_TIMEOUT):
        """
        Waits until the VM instance is ready to receive ssh connections
        """
        logger.debug(f"Waiting {self} to become ready")

        start = time.time()

        if self.public:
            self.get_public_ip()
        else:
            self.get_private_ip()

        while time.time() - start < timeout:
            if self.is_ready():
                start_time = round(time.time() - start, 2)
                logger.debug(f"{self} ready in {start_time} seconds")
                return True
            time.sleep(5)

        raise TimeoutError(f"Readiness probe expired on {self}")

    @imports.inject("ibm_cloud_sdk_core", pip_extra="ibmcloud")
    def _create_instance(ibm_cloud_sdk_core, self, user_data):
        """
        Creates a new VM instance
        """
        logger.debug("Creating new VM instance {}".format(self.name))

        security_group_identity_model = {"id": self.config["security_group_id"]}
        subnet_identity_model = {"id": self.config["subnet_id"]}
        primary_network_interface = {"name": "eth0", "subnet": subnet_identity_model, "security_groups": [security_group_identity_model]}

        boot_volume_data = {
            "capacity": self.config["boot_volume_capacity"],
            "name": f"{self.name}-{str(uuid.uuid4())[:4]}-boot",
            "profile": {"name": self.config["boot_volume_profile"]},
        }

        boot_volume_attachment = {"delete_volume_on_instance_delete": True, "volume": boot_volume_data}

        key_identity_model = {"id": self.config["ssh_key_id"]}

        instance_prototype = {}
        instance_prototype["name"] = self.name
        instance_prototype["keys"] = [key_identity_model]
        instance_prototype["profile"] = {"name": self.profile_name}
        instance_prototype["resource_group"] = {"id": self.config["resource_group_id"]}
        instance_prototype["vpc"] = {"id": self.config["vpc_id"]}
        instance_prototype["image"] = {"id": self.config["image_id"]}
        instance_prototype["zone"] = {"name": self.config["zone_name"]}
        instance_prototype["boot_volume_attachment"] = boot_volume_attachment
        instance_prototype["primary_network_interface"] = primary_network_interface

        if user_data:
            instance_prototype["user_data"] = user_data

        try:
            resp = self.vpc_cli.create_instance(instance_prototype)
        except ibm_cloud_sdk_core.ApiException as err:
            if err.code == 400 and "already exists" in err.message:
                return self.get_instance_data()
            elif err.code == 400 and "over quota" in err.message:
                logger.debug(f"Create VM instance {self.name} failed due to quota limit")
            else:
                logger.debug("Create VM instance {} failed with status code {}: {}".format(self.name, str(err.code), err.message))
            raise err

        logger.debug("VM instance {} created successfully ".format(self.name))

        return resp.result

    def _attach_floating_ip(self, fip, fip_id, instance):
        """
        Attach a floating IP address to VM
        """

        # logger.debug('Attaching floating IP {} to VM instance {}'.format(fip, instance['id']))

        # we need to check if floating ip is not attached already. if not, attach it to instance
        instance_primary_ni = instance["primary_network_interface"]

        if instance_primary_ni["primary_ipv4_address"] and instance_primary_ni["id"] == fip_id:
            # floating ip already atteched. do nothing
            logger.debug("Floating IP {} already attached to eth0".format(fip))
        else:
            self.vpc_cli.add_instance_network_interface_floating_ip(instance["id"], instance["network_interfaces"][0]["id"], fip_id)

    def _delete_floating_ip(self, fip_id):
        response = self.vpc_cli.delete_floating_ip(id=fip_id)

    def get_instance_data(self):
        """
        Returns the instance information
        """
        # if self.instance_id:
        #    self.instance_data = self.vpc_cli.get_instance(self.instance_id).get_result()

        instances_data = self.vpc_cli.list_instances(name=self.name).get_result()
        if len(instances_data["instances"]) > 0:
            self.instance_data = instances_data["instances"][0]

        return self.instance_data

    def get_instance_id(self):
        """
        Returns the instance ID
        """
        if not self.instance_id:
            instance_data = self.get_instance_data()
            if instance_data:
                self.instance_id = instance_data["id"]
            else:
                logger.debug(f"VM instance {self.name} does not exists")
        return self.instance_id

    def get_private_ip(self):
        """
        Requests the private IP address
        """
        while not self.private_ip or self.private_ip == "0.0.0.0":
            time.sleep(1)
            instance_data = self.get_instance_data()
            self.private_ip = instance_data["primary_network_interface"]["primary_ipv4_address"]

        return self.private_ip

    def get_public_ip(self):
        """
        Requests the public IP address
        """
        return self.public_ip

    def create(self, check_if_exists=False, user_data=None):
        """
        Creates a new VM instance
        """
        instance = None
        vsi_exists = True if self.instance_id else False

        if check_if_exists and not vsi_exists:
            logger.debug("Checking if VM instance {} already exists".format(self.name))
            instances_data = self.get_instance_data()
            if instances_data:
                logger.debug("VM instance {} already exists".format(self.name))
                vsi_exists = True
                self.instance_id = instances_data["id"]

        if not vsi_exists:
            instance = self._create_instance(user_data=user_data)
            self.instance_id = instance["id"]
        else:
            self.start()

        if self.public and instance:
            self._attach_floating_ip(self.public_ip, self.floating_ip_id, instance)

        return self.instance_id

    @imports.inject("ibm_cloud_sdk_core", pip_extra="ibmcloud")
    def start(ibm_cloud_sdk_core, self):
        """
        Starts the VM instance
        """
        logger.debug("Starting VM instance {}".format(self.name))

        try:
            self.vpc_cli.create_instance_action(self.instance_id, "start")
        except ibm_cloud_sdk_core.ApiException as err:
            if err.code == 404:
                pass
            else:
                raise err

        logger.debug("VM instance {} started successfully".format(self.name))

    @imports.inject("ibm_cloud_sdk_core", pip_extra="ibmcloud")
    def _delete_instance(ibm_cloud_sdk_core, self):
        """
        Deletes the VM instacne and the associated volume
        """
        logger.debug("Deleting VM instance {}".format(self.name))
        try:
            self.vpc_cli.delete_instance(self.instance_id)
        except ibm_cloud_sdk_core.ApiException as err:
            if err.code == 404:
                pass
            else:
                raise err
        self.instance_id = None
        self.private_ip = None
        self.del_ssh_client()

    @imports.inject("ibm_cloud_sdk_core", pip_extra="ibmcloud")
    def _stop_instance(ibm_cloud_sdk_core, self):
        """
        Stops the VM instance
        """
        logger.debug("Stopping VM instance {}".format(self.name))
        try:
            self.vpc_cli.create_instance_action(self.instance_id, "stop")
        except ibm_cloud_sdk_core.ApiException as err:
            if err.code == 404:
                pass
            else:
                raise err

    def stop(self):
        """
        Stops the VM instance
        """
        if self.delete_on_dismantle:
            self._delete_instance()
        else:
            self._stop_instance()

    def delete(self):
        """
        Deletes the VM instance
        """
        self._delete_instance()
        self._delete_floating_ip(self.floating_ip_id)

    def validate_capabilities(self):
        """
        Validate hardware/os requirments specified in backend config
        """
        if self.config.get("singlesocket"):
            cmd = "lscpu -p=socket|grep -v '#'"
            res = self.get_ssh_client().run_remote_command(cmd)
            sockets = set()
            for char in res:
                if char != "\n":
                    sockets.add(char)
            if len(sockets) != 1:
                raise Exception(f"Not using single CPU socket as specified, using {len(sockets)} sockets instead")


RETRIABLE = [
    "list_vpcs",
    "create_vpc",
    "get_security_group",
    "create_security_group_rule",
    "list_public_gateways",
    "create_public_gateway",
    "list_subnets",
    "create_subnet",
    "set_subnet_public_gateway",
    "list_floating_ips",
    "create_floating_ip",
    "get_instance",
    "delete_instance",
    "list_instances",
    "list_instance_network_interface_floating_ips",
    "delete_floating_ip",
    "delete_subnet",
    "delete_public_gateway",
    "delete_vpc",
    "get_instance_initialization",
    "get_key",
    "create_instance",
    "add_instance_network_interface_floating_ip",
    "get_instance",
    "create_instance_action",
    "delete_instance",
]


def decorate_instance(instance, decorator):
    for name, func in inspect.getmembers(instance, inspect.ismethod):
        if name in RETRIABLE:
            setattr(instance, name, decorator(func))
    return instance


@imports.inject("ibm_cloud_sdk_core", pip_extra="ibmcloud")
def vpc_retry_on_except(ibm_cloud_sdk_core, func):
    RETRIES = 3
    SLEEP_FACTOR = 1.5
    MAX_SLEEP = 30

    IGNORED_404_METHODS = ["delete_instance", "delete_public_gateway", "delete_vpc", "create_instance_action"]

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        sleep_time = 1

        def _sleep_or_raise(sleep_time, err):
            if i < RETRIES - 1:
                time.sleep(sleep_time)
                logger.warning((f"Got exception {err}, retrying for the {i} time, left retries {RETRIES - 1 -i}"))
                return min(sleep_time * SLEEP_FACTOR, MAX_SLEEP)
            else:
                raise err

        for i in range(RETRIES):
            try:
                return func(*args, **kwargs)
            except ibm_cloud_sdk_core.ApiException as err:
                if func.__name__ in IGNORED_404_METHODS and err.code == 404:
                    logger.debug((f"Got exception {err} when trying to invoke {func.__name__}, ignoring"))
                else:
                    sleep_time = _sleep_or_raise(sleep_time, err)
            except Exception as err:
                sleep_time = _sleep_or_raise(sleep_time, err)

    return wrapper



import uuid
from dataclasses import dataclass, field
from functools import partial

from typing import Optional, Dict, Set, List, Tuple

from skyplane import compute
from skyplane.utils import logger
from skyplane.utils.fn import do_parallel
from skyplane.utils.timer import Timer


@dataclass
class ProvisionerTask:
    """Dataclass to track a single VM provisioning job.

    :param cloud_provider: the name of the cloud provider to provision gateways
    :type cloud_provider: str
    :param region: the name of the cloud region to provision gateways
    :type region: str
    :param vm_type: the type of VMs to provision in that cloud region
    :type vm_type: str
    :param spot: whether to use the spot instances
    :type spot: bool
    :param autoterminate_minutes: terminate the VMs after this amount of time (default: None)
    :type autoterminate_minutes: int
    :param tags: a list of tags with identifications such as hostid
    :type tags: dict
    :param uuid: the uuid of the local host that requests the provision task
    :type uuid: str
    """

    cloud_provider: str
    region: str
    vm_type: str
    spot: bool = False
    autoterminate_minutes: Optional[int] = None
    tags: Dict[str, str] = field(default_factory=dict)
    uuid: str = field(default_factory=lambda: str(uuid.uuid4()))

    def __hash__(self):
        return uuid.UUID(self.uuid).int


class Provisioner:
    """Launches a single VM provisioning job."""

    def __init__(
        self,
        aws_auth: Optional[compute.AWSAuthentication] = None,
        azure_auth: Optional[compute.AzureAuthentication] = None,
        gcp_auth: Optional[compute.GCPAuthentication] = None,
        host_uuid: Optional[str] = None,
        ibmcloud_auth: Optional[compute.IBMCloudAuthentication] = None,
    ):
        """
        :param aws_auth: authentication information for aws
        :type aws_auth: compute.AWSAuthentication
        :param azure_auth: authentication information for azure
        :type azure_auth: compute.AzureAuthentication
        :param gcp_auth: authentication information for gcp
        :type gcp_auth: compute.GCPAuthentication
        :param host_uuid: the uuid of the local host that requests the provision task
        :type host_uuid: string
        :param ibmcloud_auth: authentication information for aws
        :type ibmcloud_auth: compute.IBMCloudAuthentication
        """
        self.aws_auth = aws_auth
        self.azure_auth = azure_auth
        self.gcp_auth = gcp_auth
        self.host_uuid = host_uuid
        self.ibmcloud_auth = ibmcloud_auth
        self._make_cloud_providers()
        self.temp_nodes: Set[compute.Server] = set()  # temporary area to store nodes that should be terminated upon exit
        self.pending_provisioner_tasks: List[ProvisionerTask] = []
        self.provisioned_vms: Dict[str, compute.Server] = {}

        # store GCP firewall rules to be deleted upon exit
        self.gcp_firewall_rules: Set[str] = set()

    def _make_cloud_providers(self):
        self.aws = compute.AWSCloudProvider(
            key_prefix=f"skyplane{'-'+self.host_uuid.replace('-', '') if self.host_uuid else ''}", auth=self.aws_auth
        )
        self.azure = compute.AzureCloudProvider(auth=self.azure_auth)
        self.gcp = compute.GCPCloudProvider(auth=self.gcp_auth)
        self.ibmcloud = compute.IBMCloudProvider(auth=self.ibmcloud_auth)

    def init_global(self, aws: bool = True, azure: bool = True, gcp: bool = True, ibmcloud: bool = True):
        """
        Initialize the global cloud providers by configuring with credentials

        :param aws: whether to configure aws (default: True)
        :type aws: bool
        :param azure: whether to configure azure (default: True)
        :type azure: bool
        :param gcp: whether to configure gcp (default: True)
        :type gcp: bool
        :param ibmcloud: whether to configure ibmcloud (default: True)
        :type ibmcloud: bool
        """
        logger.fs.info(f"[Provisioner.init_global] Initializing global resources for aws={aws}, azure={azure}, gcp={gcp}")
        jobs = []
        if aws:
            jobs.append(partial(self.aws.setup_global, attach_policy_arn="arn:aws:iam::aws:policy/AmazonS3FullAccess"))
        if azure:
            jobs.append(self.azure.create_ssh_key)
            jobs.append(self.azure.set_up_resource_group)
        if gcp:
            jobs.append(self.gcp.setup_global)
        if ibmcloud:
            jobs.append(self.ibmcloud.setup_global)

        do_parallel(lambda fn: fn(), jobs, spinner=False)

    def add_task(
        self,
        cloud_provider: str,
        region: str,
        vm_type: str,
        spot: bool = False,
        autoterminate_minutes: Optional[int] = None,
        tags: Optional[Dict[str, str]] = None,
    ):
        """
        Add a provision task

        :param cloud_provider: the name of the cloud provider to provision gateways
        :type cloud_provider: str
        :param region: the name of the cloud region to provision gateways
        :type region: str
        :param vm_type: the type of VMs to provision in that cloud region
        :type vm_type: str
        :param spot: whether to use the spot instances
        :type spot: bool
        :param autoterminate_minutes: terminate the VMs after this amount of time (default: None)
        :type autoterminate_minutes: int
        :param tags: a list of tags with identifications such as hostid
        :type tags: dict
        """
        if tags is None:
            tags = {"skyplane": "true", "skyplaneclientid": self.host_uuid} if self.host_uuid else {"skyplane": "true"}
        task = ProvisionerTask(cloud_provider, region, vm_type, spot, autoterminate_minutes, tags)
        self.pending_provisioner_tasks.append(task)
        logger.fs.info(f"[Provisioner.add_task] Queue {task}")
        return task.uuid

    def get_node(self, uuid: str) -> compute.Server:
        """
        Returns a provisioned VM

        :param uuid: the uuid of the provision task
        :type uuid: str
        """
        return self.provisioned_vms[uuid]

    def _provision_task(self, task: ProvisionerTask):
        with Timer() as t:
            if task.cloud_provider == "aws":
                assert self.aws.auth.enabled(), "AWS credentials not configured"
                server = self.aws.provision_instance(task.region, task.vm_type, use_spot_instances=task.spot, tags=task.tags)
            elif task.cloud_provider == "azure":
                assert self.azure.auth.enabled(), "Azure credentials not configured"
                server = self.azure.provision_instance(task.region, task.vm_type, use_spot_instances=task.spot, tags=task.tags)
            elif task.cloud_provider == "gcp":
                assert self.gcp.auth.enabled(), "GCP credentials not configured"
                server = self.gcp.provision_instance(
                    task.region,
                    task.vm_type,
                    use_spot_instances=task.spot,
                    gcp_premium_network=False,
                    tags=task.tags,
                )
            elif task.cloud_provider == "ibmcloud":
                assert self.ibmcloud.auth.enabled(), "IBM Cloud credentials not configured"
                server = self.ibmcloud.provision_instance(task.region, task.vm_type, tags=task.tags)
            else:
                raise NotImplementedError(f"Unknown provider {task.cloud_provider}")
        logger.fs.debug(f"[Provisioner._provision_task] Provisioned {server} in {t.elapsed:.2f}s")
        self.temp_nodes.add(server)
        self.provisioned_vms[task.uuid] = server
        server.wait_for_ssh_ready()
        if task.autoterminate_minutes:
            server.enable_auto_shutdown(task.autoterminate_minutes)
        self.temp_nodes.remove(server)
        return server

    def provision(self, authorize_firewall: bool = True, max_jobs: int = 16, spinner: bool = False) -> List[str]:
        """Provision the VMs in the pending_provisioner_tasks list. Returns UUIDs of provisioned VMs.

        :param authorize_firewall: whether to add authorization firewall to the remote gateways (default: True)
        :type authorize_firewall: bool
        :param max_jobs: maximum number of provision jobs to launch concurrently (default: 16)
        :type max_jobs: int
        :param spinner: whether to show the spinner during the job (default: False)
        :type spinner: bool
        """
        if not self.pending_provisioner_tasks:
            return []

        # copy list to avoid concurrency issue
        provision_tasks = list(self.pending_provisioner_tasks)
        aws_provisioned = any([task.cloud_provider == "aws" for task in provision_tasks])
        aws_regions = set([task.region for task in provision_tasks if task.cloud_provider == "aws"])
        ibmcloud_regions = set([task.region for task in provision_tasks if task.cloud_provider == "ibmcloud"])
        azure_provisioned = any([task.cloud_provider == "azure" for task in provision_tasks])
        gcp_provisioned = any([task.cloud_provider == "gcp" for task in provision_tasks])
        ibmcloud_provisioned = any([task.cloud_provider == "ibmcloud" for task in provision_tasks])

        # configure regions
        if aws_provisioned:
            do_parallel(
                self.aws.setup_region, list(set(aws_regions)), spinner=spinner, spinner_persist=False, desc="Configuring AWS regions"
            )
            logger.fs.info(f"[Provisioner.provision] Configured AWS regions {aws_regions}")

        if ibmcloud_provisioned:
            do_parallel(
                self.ibmcloud.setup_region,
                list(set(ibmcloud_regions)),
                spinner=spinner,
                spinner_persist=False,
                desc="Configuring IBM Cloud regions",
            )
            logger.fs.info(f"[Provisioner.provision] Configured IBM Cloud regions {ibmcloud_regions}")

        # provision VMs
        logger.fs.info(f"[Provisioner.provision] Provisioning {len(provision_tasks)} VMs")
        results: List[Tuple[ProvisionerTask, compute.Server]] = do_parallel(
            self._provision_task,
            provision_tasks,
            n=max_jobs,
            spinner=spinner,
            spinner_persist=spinner,
            desc="Provisioning VMs",
        )

        # configure firewall
        if authorize_firewall:
            public_ips = [s.public_ip() for _, s in results]
            # authorize access to private IPs for GCP VMs due to global VPC
            private_ips = [s.private_ip() for t, s in results if t.cloud_provider == "gcp"]

            authorize_ip_jobs = []
            if ibmcloud_provisioned:
                authorize_ip_jobs.extend([partial(self.ibmcloud.add_ips_to_security_group, r, public_ips) for r in set(ibmcloud_regions)])
            # NOTE: the following setup is for broadcast only
            if aws_provisioned:
                authorize_ip_jobs.extend([partial(self.aws.add_ips_to_security_group, r, None) for r in set(aws_regions)])
            if gcp_provisioned:

                def authorize_gcp_gateways():
                    self.gcp_firewall_rules.add(self.gcp.authorize_gateways(public_ips + private_ips))

                authorize_ip_jobs.append(authorize_gcp_gateways)

            do_parallel(
                lambda fn: fn(),
                authorize_ip_jobs,
                n=max_jobs,
                spinner=spinner,
                spinner_persist=False,
                desc="Authorizing gateways with firewalls",
            )
            logger.fs.info(f"[Provisioner.provision] Authorized AWS gateways with firewalls: {public_ips}")
            logger.fs.info(f"[Provisioner.provision] Authorized GCP gateways with firewalls: {public_ips}, {private_ips}")

        for task in provision_tasks:
            self.pending_provisioner_tasks.remove(task)

        return [task.uuid for task in provision_tasks]

    def deprovision(self, deauthorize_firewall: bool = True, max_jobs: int = 64, spinner: bool = False):
        """Deprovision all nodes. Returns UUIDs of deprovisioned VMs.

        :param deauthorize_firewall: whether to remove authorization firewall to the remote gateways (default: True)
        :type deauthorize_firewall: bool
        :param max_jobs: maximum number of deprovision jobs to launch concurrently (default: 64)
        :type max_jobs: int
        :param spinner: whether to show the spinner during the job (default: False)
        :type spinner: bool
        """
        if not self.provisioned_vms and not self.temp_nodes:
            return []

        def deprovision_gateway_instance(server: compute.Server):
            server.terminate_instance()
            idx_to_del = None
            for idx, s in list(self.provisioned_vms.items()):
                if s == server:
                    idx_to_del = idx
                    break
            if idx_to_del:
                del self.provisioned_vms[idx_to_del]
            else:
                logger.fs.warning(f"[Provisioner.deprovision] Could not find {server} in {self.provisioned_vms}")
            if server in self.temp_nodes:
                self.temp_nodes.remove(server)
            logger.fs.info(f"[Provisioner.deprovision] Terminated {server}")

        servers = list(self.provisioned_vms.values()) + list(self.temp_nodes)
        aws_deprovisioned = any([s.provider == "aws" for s in servers])
        azure_deprovisioned = any([s.provider == "azure" for s in servers])
        gcp_deprovisioned = any([s.provider == "gcp" for s in servers])
        ibmcloud_deprovisioned = any([s.provider == "ibmcloud" for s in servers])
        if azure_deprovisioned:
            logger.warning("Azure deprovisioning is very slow. Please be patient.")
        logger.fs.info(f"[Provisioner.deprovision] Deprovisioning {len(servers)} VMs")
        do_parallel(
            deprovision_gateway_instance,
            servers,
            n=max_jobs,
            spinner=spinner,
            spinner_persist=False,
            desc="Deprovisioning VMs",
        )

        # clean up firewall
        if deauthorize_firewall:
            # todo remove firewall rules for Azure
            public_ips = [s.public_ip() for s in servers]
            jobs = []
            if aws_deprovisioned:
                aws_regions = set([s.region() for s in servers if s.provider == "aws"])
                jobs.extend([partial(self.aws.remove_ips_from_security_group, r, public_ips) for r in set(aws_regions)])
                logger.fs.info(f"[Provisioner.deprovision] Deauthorizing AWS gateways with firewalls: {public_ips}")
            if gcp_deprovisioned:
                jobs.extend([partial(self.gcp.remove_gateway_rule, rule) for rule in self.gcp_firewall_rules])
                logger.fs.info(f"[Provisioner.deprovision] Deauthorizing GCP gateways with firewalls: {self.gcp_firewall_rules}")
            do_parallel(
                lambda fn: fn(), jobs, n=max_jobs, spinner=spinner, spinner_persist=False, desc="Deauthorizing gateways from firewalls"
            )



import json
import time
import uuid
from multiprocessing import BoundedSemaphore

from typing import List, Optional

from skyplane import exceptions as skyplane_exceptions
from skyplane.compute.aws.aws_auth import AWSAuthentication
from skyplane.compute.aws.aws_key_manager import AWSKeyManager
from skyplane.compute.aws.aws_network import AWSNetwork
from skyplane.compute.aws.aws_pricing import AWSPricing
from skyplane.compute.aws.aws_server import AWSServer
from skyplane.compute.cloud_provider import CloudProvider
from skyplane.utils import imports, logger
from skyplane.utils.fn import do_parallel, wait_for


class AWSCloudProvider(CloudProvider):
    pricing = AWSPricing()

    def __init__(
        self,
        key_prefix: str = "skyplane",
        auth: Optional[AWSAuthentication] = None,
        key_manager: Optional[AWSKeyManager] = None,
        network: Optional[AWSNetwork] = None,
    ):
        super().__init__()
        self.key_prefix = key_prefix
        self.auth = auth if auth else AWSAuthentication()
        self.network = network if network else AWSNetwork(self.auth)
        self.key_manager = key_manager if key_manager else AWSKeyManager(self.auth)
        self.provisioning_semaphore = BoundedSemaphore(16)

    @property
    def name(self):
        return "aws"

    @staticmethod
    def region_list() -> List[str]:
        return AWSAuthentication.get_region_config()

    @classmethod
    def get_transfer_cost(cls, src_key, dst_key, premium_tier=True):
        assert src_key.startswith("aws:")
        return cls.pricing.get_transfer_cost(src_key, dst_key, premium_tier)

    @imports.inject("botocore.exceptions", pip_extra="aws")
    def get_instance_list(exceptions, self, region: str) -> List[AWSServer]:
        ec2 = self.auth.get_boto3_resource("ec2", region)
        valid_states = ["pending", "running", "stopped", "stopping"]
        instances = ec2.instances.filter(Filters=[{"Name": "instance-state-name", "Values": valid_states}])
        try:
            instance_ids = [i.id for i in instances]
        except exceptions.ClientError as e:
            logger.error(f"error provisioning in {region}: {e}")
            return []
        return [AWSServer(f"aws:{region}", i) for i in instance_ids]

    def setup_global(self, iam_name: str = "skyplane_gateway", attach_policy_arn: Optional[str] = None):
        # Create IAM role if it doesn't exist and grant managed role if given.
        iam = self.auth.get_boto3_client("iam")
        try:
            iam.get_role(RoleName=iam_name)
        except iam.exceptions.NoSuchEntityException:
            doc = {
                "Version": "2012-10-17",
                "Statement": [{"Effect": "Allow", "Principal": {"Service": "ec2.amazonaws.com"}, "Action": "sts:AssumeRole"}],
            }
            iam.create_role(RoleName=iam_name, AssumeRolePolicyDocument=json.dumps(doc), Tags=[{"Key": "skyplane", "Value": "true"}])
        if attach_policy_arn:
            iam.attach_role_policy(RoleName=iam_name, PolicyArn=attach_policy_arn)

    def setup_region(self, region: str):
        # set up network
        self.network.make_vpc(region)
        self.network.add_ssh_to_security_group(region)

        # set up keys
        self.key_manager.ensure_key_exists(region, f"{self.key_prefix}-{region}")

    def teardown_global(self):
        def list_instance_profiles(prefix: Optional[str] = None):
            paginator = self.auth.get_boto3_client("iam").get_paginator("list_instance_profiles")
            matched_names = []
            for page in paginator.paginate():
                for profile in page["InstanceProfiles"]:
                    if prefix is None or profile["InstanceProfileName"].startswith(prefix):
                        matched_names.append(profile["InstanceProfileName"])
            return matched_names

        def delete_instance_profile(profile_name: str):
            profile = self.auth.get_boto3_resource("iam").InstanceProfile(profile_name)
            for role in profile.roles:
                profile.remove_role(RoleName=role.name)
            # delete the instance profile
            profile.delete()

        vpcs = do_parallel(self.network.get_vpcs, self.region_list(), desc="Querying VPCs", spinner=True)
        args = [(x[0], vpc.id) for x in vpcs for vpc in x[1]]
        do_parallel(lambda x: self.network.remove_sg_ips(*x), args, desc="Removing IPs from VPCs", spinner=True, spinner_persist=True)
        profiles = list_instance_profiles(prefix="skyplane-aws")
        if profiles:
            do_parallel(delete_instance_profile, profiles, desc="Deleting instance profiles", spinner=True, spinner_persist=True, n=4)

    def add_ips_to_security_group(self, aws_region: str, ips: Optional[List[str]] = None):
        """Add IPs to security group. If security group ID is None, use group named skyplane (create if not exists). If ip is None, authorize all IPs."""
        self.network.add_ips_to_security_group(aws_region, ips)

    def remove_ips_from_security_group(self, aws_region: str, ips: List[str]):
        """Remove IP from security group. If security group ID is None, return."""
        self.network.remove_ips_from_security_group(aws_region, ips)

    @imports.inject("botocore.exceptions", pip_extra="aws")
    def provision_instance(
        exceptions,
        self,
        region: str,
        instance_class: str,
        disk_size: int = 32,
        use_spot_instances: bool = False,
        name: Optional[str] = None,
        tags={"skyplane": "true"},
        aws_iam_name: str = "skyplane_gateway",
        instance_os: str = "ecs-aws-linux-2",
    ) -> AWSServer:
        assert not region.startswith("aws:"), "Region should be AWS region"
        if name is None:
            name = f"skyplane-aws-{str(uuid.uuid4().hex[:8])}"
        iam_instance_profile_name = f"{name}_profile"

        # set default image used for provisioning instances in AWS
        if instance_os == "ubuntu":
            image_id = "resolve:ssm:/aws/service/canonical/ubuntu/server/20.04/stable/current/amd64/hvm/ebs-gp2/ami-id"
        elif instance_os == "ecs-aws-linux-2":
            image_id = "resolve:ssm:/aws/service/ecs/optimized-ami/amazon-linux-2/recommended/image_id"
        else:
            raise ValueError(f"Provisioning in {region}: instance OS {instance_os} not supported")

        with self.provisioning_semaphore:
            iam = self.auth.get_boto3_client("iam", region)
            ec2 = self.auth.get_boto3_resource("ec2", region)
            ec2_client = self.auth.get_boto3_client("ec2", region)
            vpcs = self.network.get_vpcs(region)
            assert vpcs, "No VPC found"
            vpc = vpcs[0]

            # get subnet list
            def instance_class_supported(az):
                # describe_instance_type_offerings
                offerings_list = ec2_client.describe_instance_type_offerings(
                    LocationType="availability-zone", Filters=[{"Name": "location", "Values": [az]}]
                )
                offerings = [o for o in offerings_list["InstanceTypeOfferings"] if o["InstanceType"] == instance_class]
                return len(offerings) > 0

            subnets = [subnet for subnet in vpc.subnets.all() if instance_class_supported(subnet.availability_zone)]
            assert len(subnets) > 0, "No subnets found that support specified instance class"

            def check_iam_role():
                try:
                    iam.get_role(RoleName=aws_iam_name)
                    return True
                except iam.exceptions.NoSuchEntityException:
                    return False

            def check_instance_profile():
                try:
                    iam.get_instance_profile(InstanceProfileName=iam_instance_profile_name)
                    return True
                except iam.exceptions.NoSuchEntityException:
                    return False

            # wait for iam_role to be created and create instance profile
            wait_for(check_iam_role, timeout=60, interval=0.5)
            iam.create_instance_profile(InstanceProfileName=iam_instance_profile_name, Tags=[{"Key": "skyplane", "Value": "true"}])
            iam.add_role_to_instance_profile(InstanceProfileName=iam_instance_profile_name, RoleName=aws_iam_name)
            wait_for(check_instance_profile, timeout=60, interval=0.5)

            def start_instance(subnet_id: str):
                if use_spot_instances:
                    market_options = {"MarketType": "spot"}
                else:
                    market_options = {}
                return ec2.create_instances(
                    ImageId=image_id,
                    InstanceType=instance_class,
                    MinCount=1,
                    MaxCount=1,
                    KeyName=f"{self.key_prefix}-{region}",
                    TagSpecifications=[
                        {
                            "ResourceType": "instance",
                            "Tags": [{"Key": "Name", "Value": name}] + [{"Key": k, "Value": v} for k, v in tags.items()],
                        }
                    ],
                    BlockDeviceMappings=[
                        {
                            "DeviceName": "/dev/sda1",
                            "Ebs": {"DeleteOnTermination": True, "VolumeSize": disk_size, "VolumeType": "gp2"},
                        }
                    ],
                    NetworkInterfaces=[
                        {
                            "DeviceIndex": 0,
                            "Groups": [self.network.get_security_group(region).group_id],
                            "SubnetId": subnet_id,
                            "AssociatePublicIpAddress": True,
                            "DeleteOnTermination": True,
                        }
                    ],
                    IamInstanceProfile={"Name": iam_instance_profile_name},
                    InstanceInitiatedShutdownBehavior="terminate",
                    InstanceMarketOptions=market_options,
                )

            backoff = 1
            max_retries = 8
            max_backoff = 8
            current_subnet_id = 0
            for i in range(max_retries):
                try:
                    instance = start_instance(subnets[current_subnet_id].id)
                    break
                except exceptions.ClientError as e:
                    if i == max_retries - 1:
                        raise
                    elif "VcpuLimitExceeded" in str(e):
                        raise skyplane_exceptions.InsufficientVCPUException() from e
                    elif "Invalid IAM Instance Profile name" in str(e):
                        # TODO: suppress this
                        logger.fs.info(str(e))
                    elif "InsufficientInstanceCapacity" in str(e):
                        # try another subnet
                        current_subnet_id = (current_subnet_id + 1) % len(subnets)
                    else:
                        raise ValueError(f"Unexpected provisioning error: {str(e)}")
                    time.sleep(backoff)
                    backoff = min(backoff * 2, max_backoff)

            assert len(instance) == 1, f"Expected 1 instance, got {len(instance)}"
            try:
                instance[0].wait_until_running()
            except KeyboardInterrupt:
                logger.fs.warning(f"Terminating instance {instance[0].id} due to keyboard interrupt")
                instance[0].terminate()
                raise
            return AWSServer(f"aws:{region}", instance[0].id)
